from argparse import _SubParsersAction, ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from multiprocessing import Queue
import socket
from struct import unpack_from
from typing import Tuple

from ..types import AircraftState, Buttons, Controls, RunFn


def setup(subparsers: _SubParsersAction) -> Tuple[str, RunFn]:
    NAME = 'rpctask'
    parser: ArgumentParser = subparsers.add_parser(
        NAME,
        help='listen to UDP packets from Rotorcraft-Pilot Coupling Task simulator',
        formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('--ip', '-i',
                        help='listen UDP adress', default='0.0.0.0')
    parser.add_argument('--port', '-p', type=int,
                        help='listen UDP port', default=5004)

    return (NAME, run)


def run(q: Queue, args: Namespace):

    last_trim = Controls()

    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
        # Timeout is required to handle leaving with Ctrl+C
        sock.settimeout(1.0)
        sock.bind((args.ip, args.port))
        print(f'Listening for UDP packets on {args.ip}:{args.port}')

        while True:
            try:
                data, _ = sock.recvfrom(1024)
                # As named in the original FlightGear protocol file
                # cyclic has coordinates 0,0 in bottom left and 1,1 in top right
                (pitch_ctrl, roll_ctrl, pwr_ctrl,
                 pitch_target, roll_target, pwr_target,
                 cyc_ftr, coll_ftr,
                 up_bdr_cyc, low_bdr_cyc, right_bdr_cyc, left_bdr_cyc,
                 up_bdr_coll, low_bdr_coll) = unpack_from('>' + 'd' * 14, data)

                state = AircraftState()
                state.ctrl.stick_right = (roll_ctrl * 2.0) - 1.0
                state.ctrl.stick_pull = (pitch_ctrl * -2.0) + 1.0
                state.ctrl.collective_up = pwr_ctrl
                state.trgt.stick_right = (roll_target * 2.0) - 1.0
                state.trgt.stick_pull = (pitch_target * -2.0) + 1.0
                state.trgt.collective_up = pwr_target
                if cyc_ftr > 0.5:
                    state.btn |= Buttons.CYC_FTR
                if coll_ftr > 0.5:
                    state.btn |= Buttons.COLL_FTR
                state.brdr.high.stick_right = (right_bdr_cyc * 2.0) - 1.0
                state.brdr.low.stick_right = (left_bdr_cyc * 2.0) - 1.0
                state.brdr.high.stick_pull = (low_bdr_cyc * -2.0) + 1.0
                state.brdr.low.stick_pull = (up_bdr_cyc * -2.0) + 1.0
                state.brdr.high.collective_up = up_bdr_coll
                state.brdr.low.collective_up = low_bdr_coll

                if state.btn & Buttons.COLL_FTR:
                    last_trim.collective_up = state.ctrl.collective_up
                if state.btn & Buttons.CYC_FTR:
                    last_trim.stick_right = state.ctrl.stick_right
                    last_trim.stick_pull = state.ctrl.stick_pull
                state.trim = last_trim

                q.put(('smol', state.smol()))
            except socket.timeout:
                continue



from argparse import _SubParsersAction, ArgumentDefaultsHelpFormatter, ArgumentError, ArgumentParser, Namespace
from math import pi, sin
from multiprocessing import Queue
from time import sleep, time
from typing import Tuple

from ..types import AircraftState, Buttons, Controls, RunFn


def setup(subparsers: _SubParsersAction) -> Tuple[str, RunFn]:
    NAME = 'demo'
    parser: ArgumentParser = subparsers.add_parser(
        NAME,
        help='looping simulation demonstrating GUI',
        formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('--period', '-T', type=float,
                        help='loop time in seconds', default=5.0)
    parser.add_argument('--frequency', '-f', type=float,
                        help='number of messages sent per second', default=30.0)
    parser.add_argument('--no-trim', action='store_false', dest='trim',
                        help='do not demonstrate trim function')
    parser.add_argument('--no-borders', action='store_false', dest='borders',
                        help='do not demonstrate borders function')
    parser.add_argument('--outside-range', action='store_true',
                        help='demonstrate values outside allowed range')

    return (NAME, run)


def run(q: Queue, args: Namespace):
    if args.period <= 0:
        raise ArgumentError(args.frequency, 'period must be positive')
    if args.frequency <= 0:
        raise ArgumentError(args.frequency, 'frequency must be positive')

    outside_offset = 0.6 if args.outside_range else 0.0

    def val(phase):
        """Generate a sinusoidal value with phase offset (from 0 to 1)"""
        return 0.6 * sin((time() / args.period + phase) * 2 * pi) + outside_offset

    def cycle_index(): return int(time() // args.period) % 3
    def current_phase(): return (time() / args.period) % 1

    last_trim = Controls()

    while True:
        state = AircraftState()

        state.ctrl.stick_pull = val(0.1)
        state.ctrl.stick_right = val(0.35)
        state.ctrl.collective_up = 0.3 + 0.5 * (val(0.2) + outside_offset)

        state.trgt.stick_pull = val(0.55)
        state.trgt.stick_right = val(0.3)
        state.trgt.collective_up = 0.3 + 0.5 * (val(0.4) + outside_offset)

        if cycle_index() == 1 and args.trim:
            if current_phase() < 0.4:
                state.btn |= Buttons.COLL_FTR
            if current_phase() > 0.6:
                state.btn |= Buttons.CYC_FTR

        if state.btn & Buttons.COLL_FTR:
            last_trim.collective_up = state.ctrl.collective_up
        if state.btn & Buttons.CYC_FTR:
            last_trim.stick_right = state.ctrl.stick_right
            last_trim.stick_pull = state.ctrl.stick_pull
        state.trim = last_trim

        if cycle_index() == 2 and args.borders:
            if current_phase() < 0.5:
                state.brdr.low = Controls(-0.8, -0.8, 0.1, -0.8, 0.1)
                state.brdr.high = Controls(0.8, 0.8, 0.9, 0.8, 0.9)
            if current_phase() > 0.5:
                state.brdr.low = Controls(-0.5, -0.5, 0.25, -0.5, 0.25)
                state.brdr.high = Controls(0.5, 0.5, 0.75, 0.5, 0.75)

        q.put(('smol', state.smol()))
        sleep(1 / args.frequency)



import imp
import queue
import eventlet
from multiprocessing import Queue
from os import path
import socketio


def background_loop(q: Queue, sio: socketio.Server):
    while True:
        try:
            while not q.empty():
                event, data = q.get()
                sio.emit(event, data)
        except queue.Empty:
            pass
        eventlet.sleep(0.02)


def run_server(q: Queue, host: str, port: int):
    # specifying just local path breaks when run as a module
    root_path = path.abspath(path.dirname(__file__))
    static_files = {
        '/': {'content_type': 'text/html', 'filename': path.join(root_path, 'index.html')},
        '/static': path.join(root_path, 'static'),
    }

    sio = socketio.Server()
    app = socketio.WSGIApp(sio, static_files=static_files)
    eventlet.spawn(background_loop, q, sio)
    eventlet.wsgi.server(eventlet.listen((host, port)), app)



from argparse import _SubParsersAction, Namespace
from dataclasses import dataclass
from enum import IntFlag
from multiprocessing import Queue
from typing import Callable, List, NoReturn, Tuple


RunFn = Callable[[Queue, Namespace], NoReturn]
SetupFn = Callable[[_SubParsersAction], Tuple[str, RunFn]]


@dataclass
class Controls:
    stick_right: float = 0
    stick_pull: float = 0
    throttle: float = 0
    pedals_right: float = 0
    collective_up: float = 0

    def smol(self) -> List[float]:
        return [
            self.stick_right,
            self.stick_pull,
            self.throttle,
            self.pedals_right,
            self.collective_up
        ]


class Borders:
    def __init__(self) -> None:
        self.low = Controls(-1, -1, 0, -1, 0)
        self.high = Controls(1, 1, 1, 1, 1)

    def smol(self) -> dict:
        return {
            'low': self.low.smol(),
            'high': self.high.smol()
        }


class Buttons(IntFlag):
    NONE = 0
    CYC_FTR = 1 << 0
    COLL_FTR = 1 << 1

    def smol(self) -> List[int]:
        return [int(self)]


class AircraftState:
    """Full state of displayed aircraft initialised with defaults"""

    def __init__(self) -> None:
        self.ctrl = Controls()
        """Current control inceptors position"""
        self.trgt = Controls()
        """Target inceptors position"""
        self.trim = Controls()
        """Controls trim"""
        self.brdr = Borders()
        """Task borders for inceptors"""
        self.btn = Buttons.NONE
        """Currently pressed buttons"""

    def smol(self) -> dict:
        """Return self as dictionary with SMOL-defined keys"""
        d = dict()
        for key in ['ctrl', 'trgt', 'trim', 'brdr', 'btn']:
            d[key] = getattr(self, key).smol()
        return d



import argparse
from multiprocessing import Process, Queue
from typing import Dict

from .server import run_server
from .types import RunFn, SetupFn

from .sources import demo, rpctask


def main():
    parser = argparse.ArgumentParser(
        prog='lidia',
        description='serve an aircraft instruments panel as a web page',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog='you can also see help for a specific source: "lidia <src> --help"')
    parser.add_argument('--http-host', '-H', type=str,
                        help='hosts to accept for web page', default='0.0.0.0')
    parser.add_argument('--http-port', '-P', type=int,
                        help='port to serve the web page on', default=5555)
    subparsers = parser.add_subparsers(title='source', required=True, dest='source',
                                       help='source name', description='select where to get aircraft state')

    sources: Dict[str, RunFn] = {}
    for source_module in [demo, rpctask]:
        setup: SetupFn = source_module.setup
        name, run_function = setup(subparsers)
        sources[name] = run_function

    args = parser.parse_args()

    queue = Queue()
    server_process = Process(target=run_server, args=(
        queue, args.http_host, args.http_port))
    server_process.start()

    print(f'Lidia GUI for {args.source} on http://localhost:{args.http_port}')
    try:
        (sources[args.source])(queue, args)

    except KeyboardInterrupt:
        print('Exiting main loop')

    server_process.terminate()


if __name__ == '__main__':
    main()
